﻿using System;
using System.Windows.Forms;
using System.Threading.Tasks;
using iSDK;
using System.ComponentModel;
using System.IO;
using System.Collections.Generic;
using CanonController.Properties;
using System.Diagnostics;

namespace CanonController
{
    public partial class frmLensControl : Form
    {
        string PORT = "";
        iSDKLC myLens = new iSDKLC();
        Stopwatch FocusRateLimiter = new Stopwatch();
        Stopwatch IrisRateLimiter = new Stopwatch();
        int FocusLastUpdateValue;
        double IrisLastUpdateValue;

        public frmLensControl()
        {

            InitializeComponent();
            this.Text = "illunis Canon Lens Controller";

            //Disble buttons until connect and initialize
            btnDisconnect.Enabled = false;
            btnConnect.Enabled = true;
            gbFocus.Enabled = false;
            gbIris.Enabled = false;
            btnController.Enabled = false;
            update_combo();
            SetupToolTips();

        }

        private void LensAttachDetach(object sender, EventArgs e)
        {
            if (myLens.LensPresent())
            {
                LensAttached();
            }
            else
            {
                LensDetached();
            }
        }

        private void LensAttached()
        {
            txtLensName.SafeInvoke(d => d.Text = myLens.GetLensName());
            this.SafeInvoke(m => m.SetupFocus());
            this.SafeInvoke(m => m.SetupIris());
        }

        private void LensDetached()
        {
            txtLensName.SafeInvoke(d => d.Text = "Lens Not Detected");
            gbFocus.SafeInvoke(d => d.Enabled = false);
            gbIris.SafeInvoke(d => d.Enabled = false);
        }

        private void SetupToolTips()
        {
            toolTip.SetToolTip(this.btnSaveFocus, "Saves current focus to EEPROM");
            toolTip.SetToolTip(this.btnRestoreFocus, "Restores saved focus position");
            toolTip.SetToolTip(this.cbCalFocus, "Calibrates focus each time the serial port is connected");
            toolTip.SetToolTip(this.cbLensCheck, "Polls lens state periodically");
            toolTip.SetToolTip(this.btnFocusRefresh, "Updates form with current focus values");
            toolTip.SetToolTip(this.btnIrisRefresh, "Updates form with current aperature values");
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            int waitreturn = 0;
            DialogResult result;

        retry:
            Cursor.Current = Cursors.WaitCursor;
            myLens.LensPresenceChanged += LensAttachDetach; //subscribe to lens precence change event 
            waitreturn = myLens.PortOpen(PORT);
            if (waitreturn == -1)
            {
                result = MessageBox.Show("Initialize Port Failed", " Retry?", MessageBoxButtons.YesNo);
                if (result == System.Windows.Forms.DialogResult.No)
                {
                    // Closes the parent form.
                    System.Windows.Forms.Application.Exit();
                }
                else if (result == DialogResult.Yes)
                {
                    Cursor.Current = Cursors.Default;
                    goto retry;
                }

            }
            else
            {


                btnDisconnect.Enabled = true;
                btnConnect.Enabled = false;
                btnReadLens.Enabled = true;
                btnController.Enabled = true;
                btnSend.Enabled = true;
                Cursor.Current = Cursors.Default;
                myLens.LensHeartbeat(cbLensCheck.Checked); //enables/disables SDK lens heartbeat
            }

        }

        private void btnDisconnect_Click(object sender, EventArgs e)
        {
            myLens.LensPresenceChanged -= LensAttachDetach; //unsubscribe to lens precence change event

            //disable buttons
            btnDisconnect.Enabled = false;
            btnReadLens.Enabled = false;
            btnConnect.Enabled = true;
            gbFocus.Enabled = false;
            gbIris.Enabled = false;
            btnController.Enabled = false;
            btnSend.Enabled = false;
            myLens.PortClose();
        }

        private void frmLensControl_FormClosing(object sender, FormClosingEventArgs e)
        {
            myLens.PortClose();
            Settings.Default.Save(); //save heartbeat preference for next app run
        }
        private void SetupFocus()
        {
            int far = 0;
            int near = 0;
            int current = 0;
            Cursor.Current = Cursors.WaitCursor;

            far = myLens.GetFocusFar();
            near = myLens.GetFocusNear();
            current = myLens.GetFocusCurrent();
            label1.Text = far.ToString();

            // This is possible due to the encoder uncertainty
            if (current > far) far = current;

            label1.Text = far.ToString();
            label4.Text = near.ToString();

            trackFocus.Minimum = near;
            trackFocus.Maximum = far;
            trackFocus.Value = current;
            FocusLastUpdateValue = current;
            lblFocusValue.Text = trackFocus.Value.ToString();
            trackFocus.TickFrequency = (far - near) / 20;
            gbFocus.Enabled = true;
            FocusRateLimiter.Start();

            Cursor.Current = Cursors.Default;
        }

        private void SetupIris()
        {
            double dmin = 0;
            double dmax = 0;
            double dcurrent = 0;

            Cursor.Current = Cursors.WaitCursor;

            dmin = myLens.GetIrisMin();
            dmax = myLens.GetIrisMax();
            dcurrent = myLens.GetIrisCurrent();
            label14.Text = dmin.ToString();
            label11.Text = dmax.ToString();
            lblIrisValue.Text = dcurrent.ToString();
            trackIris.Minimum = (int)(dmin * (double)10.0);
            trackIris.Maximum = (int)(dmax * (double)10.0);
            trackIris.Value = (int)(dcurrent * (double)10.0);
            IrisLastUpdateValue = trackIris.Value;
            trackIris.TickFrequency = (trackIris.Maximum - trackIris.Minimum) / 20;
            gbIris.Enabled = true;
            IrisRateLimiter.Start();

            Cursor.Current = Cursors.Default;
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            btnDisconnect_Click(sender, e);
            PORT = comboBox1.SelectedItem.ToString();
        }

        private void BtnReadLens_Click(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;
            textBox1.Text = myLens.GetLensStatus();
            textBox1.Refresh();
            Cursor.Current = Cursors.Default;
        }

        private void btnFocusNear_Click(object sender, EventArgs e)
        {
            if (trackFocus.Value != trackFocus.Minimum)
            {
                trackFocus.Value--;
                lblFocusValue.Text = trackFocus.Value.ToString();
                myLens.SetFocusAbsolute(trackFocus.Value);
            }
        }

        private void btnFocusFar_Click(object sender, EventArgs e)
        {
            if (trackFocus.Value != trackFocus.Maximum)
            {
                trackFocus.Value++;
                lblFocusValue.Text = trackFocus.Value.ToString();
                myLens.SetFocusAbsolute(trackFocus.Value);

            }
        }

        private void btnIrisOpen_Click(object sender, EventArgs e)
        {
            if (trackIris.Value != trackIris.Minimum)
            {
                trackIris.Value--;
                trackIris_MouseUp(sender, null);

            }
        }

        private void btnIrisClose_Click(object sender, EventArgs e)
        {
            if (trackIris.Value != trackIris.Maximum)
            {
                trackIris.Value++;
                trackIris_MouseUp(sender, null);
            }
        }

        private void btnController_Click(object sender, EventArgs e)
        {
            //reads lens controller firmware version
            textBox1.Clear();
            textBox1.Text = myLens.GetVersion();
        }

        private void btnSend_Click(object sender, EventArgs e)
        {
            //sends commands to lens controller
            Cursor.Current = Cursors.WaitCursor;
            textBox1.Text = myLens.PortWrite(txtSend.Text);
            Cursor.Current = Cursors.Default;
        }

        private void commandHelpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //sends controller ? command (lists valid controller commands) and displays response in new window
            String HelpText = myLens.PortWrite("?");
            Form frmHelp = new frmHelp(HelpText);
            frmHelp.Show();
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            //updates serial port list
            update_combo();
        }

        private void update_combo()
        {
            //show list of valid com ports
            comboBox1.Items.Clear();
            foreach (string s in System.IO.Ports.SerialPort.GetPortNames())
            {
                comboBox1.Items.Add(s);
            }
            //show the first entry
            comboBox1.SelectedIndex = 0;
            if (comboBox1.Items.Count == 1)
                PORT = comboBox1.SelectedItem.ToString();
        }

        //adjusts the focus as the slider is moved. this is rate limited to 1 update every 50ms
        private void trackFocus_Scroll(object sender, EventArgs e)
        {
            if (trackFocus.Value != FocusLastUpdateValue && FocusRateLimiter.ElapsedMilliseconds > 50)
            {
                lblFocusValue.Text = trackFocus.Value.ToString();
                myLens.SetFocusAbsolute(trackFocus.Value);
                FocusLastUpdateValue = trackFocus.Value;
                FocusRateLimiter.Restart();
            }
        }

        //adjusts the iris as the slider is moved. this is rate limited to 1 update every 50ms
        private void trackIris_Scroll(object sender, EventArgs e)
        {
            if (trackIris.Value != IrisLastUpdateValue && IrisRateLimiter.ElapsedMilliseconds > 50)
            {
                myLens.SetIrisAbsolute(trackIris.Value / 10.0);
                lblIrisValue.Text = myLens.GetIrisCurrent().ToString(); //lens can only do 1/8 fstop this reads back the acctual set value
                IrisLastUpdateValue = trackIris.Value;
                IrisRateLimiter.Restart();
            }
        }

        //if updates the focus to the final slider position after mouse button release
        private void trackFocus_MouseUp(object sender, MouseEventArgs e)
        {
            if (trackFocus.Value != FocusLastUpdateValue)
            {
                lblFocusValue.Text = trackFocus.Value.ToString();
                myLens.SetFocusAbsolute(trackFocus.Value);
                FocusLastUpdateValue = trackFocus.Value;
            }

        }

        //if updates the iris to the final slider position after mouse button release
        private void trackIris_MouseUp(object sender, MouseEventArgs e)
        {
            if (trackIris.Value != IrisLastUpdateValue)
            {
                myLens.SetIrisAbsolute(trackIris.Value / 10.0);
                lblIrisValue.Text = myLens.GetIrisCurrent().ToString(); //lens can only do 1/8 fstop this reads back the acctual set value
                IrisLastUpdateValue = trackIris.Value;
            }

        }

        private void loadLensNamesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //loads a lens ID/Lens name list into the Lens Controller EEPROM for lens identification
            //this is list will already be loaded at illunis. 
            //Contact illunis if you require the correctly formatted list for modification


            UInt16 LensID;
            UInt16 LensIndex;
            UInt16 UpperLensIndex = 256; //starting index for LensID > 255
            String LensName;
            List<string> SerialList = new List<string>();
            OpenFileDialog LensFile = new OpenFileDialog();
            LensFile.Filter = "txt file|*.txt|All|*.*";
            LensFile.Title = "Choose Lens List file";

            if (LensFile.ShowDialog() != DialogResult.OK)
                return;

            Cursor.Current = Cursors.WaitCursor;
            var LensLines = File.ReadAllLines(LensFile.FileName);

            foreach (string line in LensLines)
            {
                var parts = line.Split('=');
                if (UInt16.TryParse(parts[0],out LensID)) //filters out negative and x.x numbers from list
                {
                    LensIndex = (LensID < 256) ? LensID : UpperLensIndex++;
                    LensName = parts[1].Substring(0, parts[1].Contains(" or ") ? parts[1].IndexOf(" or ") : parts[1].Length); //trims off any alternate lens names
                    if (LensID != 65535) //skip N/A LensID
                        SerialList.Add("sl " + LensIndex.ToString() + " " + LensID.ToString() + LensName);
                }
            }

            SerialList.Add("sl 0 0 Name Table Initialized"); //last entry confirms table was loaded to controller

            frmLoadProgress LoadProgress = new frmLoadProgress(myLens, SerialList);
            LoadProgress.Left = this.Left;
            LoadProgress.Top = (this.Bottom - this.Top)/2;

            LoadProgress.ShowDialog(); //actual load is done in a seperate progress window

            Cursor.Current = Cursors.Default;
        }

        private void cbLensCheck_CheckedChanged(object sender, EventArgs e)
        {
            //enables and disables the ISDK lens prescense heartbeat
            myLens.LensHeartbeat(cbLensCheck.Checked);
            btnLensCheck.Visible = !cbLensCheck.Checked;
        }

        private void btnLensCheck_Click(object sender, EventArgs e)
        {
            //checks if a lens is attached. When LensHeartbeat is disabled, this will update myLens.LensPresent() 
            //and trigger the myLens.LensPresenceChanged event if there is a change
            myLens.LensPresent();
        }

        private void btnFocusRefresh_Click(object sender, EventArgs e)
        {
            //only needed if a generic command send is used to change focus settings
            SetupFocus();
        }

        private void btnIrisRefresh_Click(object sender, EventArgs e)
        {
            //only needed if a generic command send is used to change iris settings
            SetupIris();
        }

        private void btnCalFocus_CheckedChanged(object sender, EventArgs e)
        {
            //enables and disables calibrating the focus range each time a lens is attached
            myLens.FocusCalibrationOnConnect(cbCalFocus.Checked);

        }

        private void controllerTestToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (btnSend.Enabled && myLens.LensPresent())  //port is open and lens present
            {
                frmControllerTest TestForm = new frmControllerTest(myLens);
                TestForm.Left = this.Left;
                TestForm.Top = (this.Bottom - this.Top) / 2;
                TestForm.ShowDialog();
            }
            else
            {
                MessageBox.Show("Port must be open and lens connected to perform test.", "Can not perform Test.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void btnSaveFocus_Click(object sender, EventArgs e)
        {
            //saved current focus position to EEPROM
            myLens.SaveFocusState();
        }

        private void btnRestoreFocus_Click(object sender, EventArgs e)
        {
            Cursor = Cursors.WaitCursor;
            //calibrates then sets current focus to position saved in EEPROM
            myLens.RestoreFocusSatate();
            System.Threading.Thread.Sleep(4000); //lens can take up to 4 seconds to restore focus
            SetupFocus(); //update track bar endpoints and current value
            Cursor = Cursors.Default;

        }


    }


    // This class is used to update the form when the myLens.LensPresenceChanged event is triggered
    public static class Extensions
    {
        public static TResult SafeInvoke<T, TResult>(this T isi, Func<T, TResult> call) where T : ISynchronizeInvoke
        {
            if (isi.InvokeRequired)
            {
                return (TResult)isi.Invoke(call, new object[] { isi });
            }
            else
                return call(isi);
        }

        public static void SafeInvoke<T>(this T isi, Action<T> call) where T : ISynchronizeInvoke
        {
            if (isi.InvokeRequired) isi.BeginInvoke(call, new object[] { isi });
            else
                call(isi);
        }
    }
}
